# tcwindprofile/tc_outer_windprofile.py

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%% Wind profile from R0mean --> R34kt
#%% Exact solution to model of non-convecting wind profile (Emanuel 2004; Chavas et al. 2015 JAS)
#%% Notes:
#%% - It is a single integration inwards from R0. *Cannot* integrate out
#%%   from R34kt, it may blow up (math is weird)
#%% - Could do a simpler approximate profile instead, but this solution is
#%%   just as fast doing some sort of curve fit, so might as well use exact

import numpy as np

def E04_outerwind_r0input_nondim_MM0(r0, fcor, Cd, w_cool, Nr):
    """
    Exact Emanuel 2004 inner integration from R0→R34kt
    Returns:
      rrfracr0 : nondimensional radius array (r/r0)
      MMfracM0 : nondimensional angular momentum (M/M0)
    """
    fcor = abs(fcor)
    M0 = 0.5 * fcor * r0**2
    dr = 0.001 if (200e3 < r0 < 2500e3) else 0.0001
    Nr = min(Nr, int(1/dr))
    rrfracr0 = np.arange(1 - (Nr-1)*dr, 1+dr, dr)
    MM = np.full(rrfracr0.shape, np.nan)
    MM[-1] = 1
    rtemp, Mtemp = rrfracr0[-2], MM[-1]
    MM[-2] = Mtemp

    # loop inward
    for i in range(Nr-2):
        gam = Cd * fcor * r0 / w_cool
        dM = gam * ((Mtemp - rtemp**2)**2) / (1 - rtemp**2)
        Mtemp -= dM * dr
        rtemp -= dr
        MM[-3-i] = Mtemp

    return rrfracr0, MM


def outer_windprofile(r0_m, fcor, Cd, w_cool, V34kt_ms):
    """
    Wrapper that:
      1) calls E04_outerwind_r0input_nondim_MM0 to get rrfracr0, MM
      2) builds a physical wind profile vv_E04approx(r)
      3) zooms to r < 1.2 R0 and interpolates vv(r)
      4) profile set to NaN inside of radius of input wind speed (e.g. V34kt_ms))
    Returns:
      rr_m : radii [m]
      vv_outer_ms : wind speeds [m/s]
    """
    # 1) exact integration
    Nr=100000
    rrfracr0_E04, MMfracM0_E04 = E04_outerwind_r0input_nondim_MM0(r0_m, fcor, Cd, w_cool, Nr)
    M0_E04approx = 0.5 * fcor * r0_m**2
    rr_E04approx = rrfracr0_E04 * r0_m
    vv_E04approx = (M0_E04approx / r0_m) * ((MMfracM0_E04 / rrfracr0_E04) - rrfracr0_E04)
    vv_E04approx[vv_E04approx > 2 * V34kt_ms] = np.nan
    
    # Zoom into relevant radii
    r0_plot = 1.2 * r0_m
    dr = 100     #[m]; radial increment
    rr_m = np.arange(0, r0_plot + dr, dr)  # [m]
    # Interpolate approx-E04 solution to original radius vector
    vv_outer_ms = np.interp(rr_m, rr_E04approx, vv_E04approx)

    # # 2) build velocity
    # M0 = 0.5 * fcor * r0_m**2
    # rr_E04 = rrfracr0 * r0_m
    # vv = (M0 / r0_m) * ((MMfracM0 / rrfracr0) - rrfracr0)
    # vv[vv > 2 * V34kt_ms] = np.nan

    # # 3) zoom & interpolate
    # r0_plot = 1.2 * r0_m
    # rr_m= rr_full_m[rr_full_m < r0_plot]
    # vv_interp_ms = np.interp(rr_m, rr_E04, vv)
    return rr_m, vv_outer_ms



import numpy as np
import matplotlib.pyplot as plt

def generate_wind_profile(Vmaxmean_ms, Rmax_km, R34ktmean_km, lat, plot=False):
    """
    Model described in Tao et al. 2025 GRL

    Analytic wind profile combining segments for:
    1) eye: r<Rmax (linear model);
    2) inner-core: Rmax to R34kt (linear-M model; Tao+ 2023 GRL);
    3) intermediate radii: R34kt to transition radius (modified Rankine model; Tao+ 2023 GRL, Klotzbach+ 2022 JGRA); and
    4) large radii: transition radius to outer radius (Ekman suction model; Emanuel 2004; Chavas+ 2015/2016 JAS).

    Inputs:
    - Vmaxmean_ms: Maximum wind speed at Rmax [m/s]
    - Rmax_km: Radius of maximum wind [km]
    - R34ktmean_km: Radius of 34-kt wind [km]
    - lat: Latitude [degrees]
    - plot: Optional flag to plot the profile

    Returns:
    - rr_km_interp: Radius vector [km]
    - vv_ms_interp: Wind speed profile [m/s]
    - R0_km: Outer radius where wind goes to zero [km]
    - Raa_km, Rba_km: Transition radii [km]
    - ir0, ira, irb: Indices of transitions (SL→mR, mR→E04, outer edge)
    """

    # Constants
    ms_kt = 0.5144444
    V34kt_ms = 34 * ms_kt
    omega = 7.2921e-5   #[s^{-1]}]; Earth's rotation rate
    fcor = np.abs(2 * omega * np.sin(np.radians(lat)))  #[s^{-1]}]; absolute value of Coriolis parameter at storm center latitude

    # Convert to meters
    Rmax_m = Rmax_km * 1000
    R34ktmean_m = R34ktmean_km * 1000
    dR = 100

    # Angular momentum
    Mmax = Vmaxmean_ms * Rmax_m + 0.5 * fcor * Rmax_m**2
    M34kt = V34kt_ms * R34ktmean_m + 0.5 * fcor * R34ktmean_m**2
    SL34kt = (M34kt / Mmax - 1) / (R34ktmean_m / Rmax_m - 1)

    Aa = 0.5 / fcor * (SL34kt * Mmax / Rmax_m)**2 + Mmax * (1 - SL34kt)
    Raa_m = SL34kt * Mmax / fcor / Rmax_m
    sqrt_term = np.sqrt(2 * Aa / fcor)
    chi = 1.5 #[-]; physical outer profile slope parameter; set to 1.5 constant (see Tao+ 2025, forthcoming); (= 2*C_d/w_cool)
    Rba_m = 0.5 * np.sqrt(chi * Aa * sqrt_term + Aa / (2 * fcor)) - 0.25 * sqrt_term
    R0a_m = 0.5 * np.sqrt(chi * Aa * sqrt_term + Aa / (2 * fcor)) + 0.75 * sqrt_term

    # Radius vector
    rr_full = np.arange(0, R0a_m + dR, dR)

    # Wind profiles
    V_eye = Vmaxmean_ms * (rr_full / Rmax_m)
    V_eye[rr_full > Rmax_m] = 0

    V_SL = ((SL34kt * (rr_full / Rmax_m - 1) + 1) * Mmax - 0.5 * fcor * rr_full**2) / rr_full
    alpha = 1.0
    V_mR = Aa * rr_full**(-alpha)
    V_E04 = (Aa - 0.5 * fcor * Rba_m**2) / rr_full + fcor * Rba_m - 0.5 * fcor * rr_full

    # Region indices
    ira = np.where(rr_full <= Raa_m)[0][-1]
    irb = np.where(rr_full <= Rba_m)[0][-1]
    ir0 = np.where(rr_full <= R0a_m)[0][-1]

    # Composite profile
    vv_ms_interp = np.zeros_like(rr_full)
    vv_ms_interp[rr_full < Rmax_m] = V_eye[rr_full < Rmax_m]
    vv_ms_interp[(rr_full >= Rmax_m) & (rr_full <= Raa_m)] = V_SL[(rr_full >= Rmax_m) & (rr_full <= Raa_m)]
    vv_ms_interp[(rr_full > Raa_m) & (rr_full <= Rba_m)] = V_mR[(rr_full > Raa_m) & (rr_full <= Rba_m)]
    vv_ms_interp[(rr_full > Rba_m) & (rr_full <= R0a_m)] = V_E04[(rr_full > Rba_m) & (rr_full <= R0a_m)]

    # Convert outputs to km
    rr_km_interp = rr_full / 1000
    R0_km = R0a_m / 1000
    Raa_km = Raa_m / 1000
    Rba_km = Rba_m / 1000

    # Optional plot
    if plot:
        fig, ax = plt.subplots(figsize=(15/2.54, 15/2.54))
        ax.plot(rr_km_interp, vv_ms_interp, 'm-', linewidth=3)
        # ax.axvline(Raa_km, linestyle='--', color='b', label='Raa')
        # ax.axvline(Rba_km, linestyle='--', color='r', label='Rba')
        # ax.axvline(R0_km, linestyle='--', color='g', label='R0')
        ax.plot(Rmax_km, Vmaxmean_ms, 'ko', label='(Rmax, Vmax)')
        ax.plot(R34ktmean_km, V34kt_ms, 'ko', label='(R34kt, 34 kt)')
        ax.set_xlabel('Radius [km]')
        ax.set_ylabel('Azimuthal wind speed [m/s]')
        ax.set_title('Complete wind profile')
        ax.set_xlim(0, R0_km * 1.1)
        ax.set_ylim(0, Vmaxmean_ms * 1.1)
        ax.grid(True)
        # ax.legend()
        annotation = (f"Inputs:\n"
                      f"Vmax_mean = {Vmaxmean_ms:.1f} m/s\n"
                      f"Rmax = {Rmax_km:.1f} km\n"
                      f"(R34kt_mean, V34kt_ms) = ({R34ktmean_km:.1f} km, {V34kt_ms:.1f} m/s)\n"
                      f"lat = {lat:.1f}°N\n\n"
                      f"Output:\n"
                      f"R0_mean = {R0_km:.1f} km")
        ax.text(0.95, 0.95, annotation, transform=ax.transAxes, ha='right', va='top',
                fontsize=10, bbox=dict(facecolor='white', edgecolor='none', alpha=0.7))
        plt.savefig('windprofile.jpg', format='jpeg')
        plt.show()
        print("Made plot of wind profile!")

    return rr_km_interp, vv_ms_interp, R0_km



# windprofile.py

## Create a fast and robust radial profile of the tropical cyclone rotating wind from inputs Vmax, R34kt, optional Rmax, and latitude.


# tcwindprofile/full_profile.py

from typing import Optional

def run_full_wind_model(
    VmaxNHC_kt: float,
    Vtrans_kt: float,
    R34kt_quad_max_nautmi: float,
    lat: float,
    Penv_mb: float,
    plot: bool = False,
    Rmax_km: Optional[float] = None
):
    """
    Full modeling pipeline:
    - If no Rmax input: estimate Rmax from R34kt -- ref Chavas and Knaff 2022 WAF
    - Estimate R0 from R34kt: approximate version of outer model ref Emanuel 2004 / Chavas et al. 2015 JAS / Chavas and Lin 2016 JAS
    - Generate wind profile: merge simple inner + outer models, ref Klotzbach et al. 2022 JGR-A / Chavas and Lin 2016 JAS
    - Estimate Pmin: ref Chavas Knaff Klotzbach 2025 WAF
    - Generate pressure profile that matches Pmin: ref Chavas Knaff Klotzbach 2025 WAF
    """
    # Convert input units
    ms_per_kt = 0.5144444
    km_per_nautmi = 1.852
    VmaxNHC_ms = VmaxNHC_kt * ms_per_kt
    Vtrans_ms = Vtrans_kt * ms_per_kt
    Vmaxmean_ms = VmaxNHC_ms - 0.55 * Vtrans_ms  #Lin and Chavas (2012), Chavas Knaff Klotzbach (2025)
    R34ktmean_km = 0.85 * R34kt_quad_max_nautmi * km_per_nautmi
    
    V34kt_ms = 17.49
    
    if Vmaxmean_ms < V34kt_ms:
        raise ValueError("Vmaxmean_ms cannot be < 34 kt")
    ###########################
    # 1) Estimate Rmax (CK22) unless provided by user
    if Rmax_km is None:
        from tcwindprofile.tc_rmax_estimatefromR34kt import predict_Rmax_from_R34kt
        Rmax_km = predict_Rmax_from_R34kt(
            VmaxNHC_ms=VmaxNHC_ms,
            R34ktmean_km=R34ktmean_km,
            lat=lat
        )

    ###########################
    # 2) Estimate wind profile + R0 (Tao et al. 2025 GRL; approximation to CLE15 model)
    # print(Vmaxmean_ms,Rmax_km,R34ktmean_km,lat,plot)
    from tcwindprofile.windprofile import generate_wind_profile
    # from tcwindprofile.tc_outer_radius_estimate import estimate_outer_radius
    # from tcwindprofile.tc_outer_windprofile import outer_windprofile
    
    rr_km, vv_ms, R0_km = generate_wind_profile(
        Vmaxmean_ms=Vmaxmean_ms,
        Rmax_km=Rmax_km,
        R34ktmean_km=R34ktmean_km,
        lat=lat,
        # plot=True
        plot=True
    )

    ###########################
    # 3) Estimate Pmin (CKK25)
    # print(VmaxNHC_ms,R34ktmean_km,lat,Vtrans_ms,Penv_mb)
    from tcwindprofile.tc_pmin_estimatefromR34kt import predict_Pmin_from_R34kt
    
    Pmin_estimate_mb, dP_estimate_mb = predict_Pmin_from_R34kt(
        VmaxNHC_ms=VmaxNHC_ms,
        R34ktmean_km=R34ktmean_km,
        lat=lat,
        Vtrans_ms=Vtrans_ms,
        Penv_mb=Penv_mb
    )

    ###########################
    # 4) Calculate pressure profile from wind field and matches Pmin
    from tcwindprofile.pressure_profile import pressure_profile_calcfromwindprofile                        # [deg]
    
    pp_mb = pressure_profile_calcfromwindprofile(
        rr_km=rr_km,
        vv_ms=vv_ms,
        Renv_km=R0_km,
        Penv_mb=Penv_mb,
        Pmin_mb=Pmin_estimate_mb,
        lat=lat,
        plot=False)
    
    ###########################
    # 5) Return final results
    return {
        "rr_km": rr_km,
        "vv_ms": vv_ms,
        "pp_mb": pp_mb,
        "Vmaxmean_ms": Vmaxmean_ms,
        "Rmax_km": Rmax_km,
        "V34kt_ms": V34kt_ms,
        "R34ktmean_km": R34ktmean_km,
        "R0_km": R0_km,
        "lat": lat,
        "Penv_mb": Penv_mb,
        "Pmin_mb": Pmin_estimate_mb,
    }


# test_windprofile_all.py

## Create a fast and robust radial profile of the tropical cyclone rotating wind from inputs Vmax, R34kt, latitude, and Vtrans.


import os
import sys

# Add parent directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import numpy as np
import math

############################################################
# NHC/Best Track Operational Inputs
VmaxNHC_kt = 97.19  #100 [kt]; NHC storm intensity (point-max wind speed)
Vtrans_kt = 0    #20 [kt]; storm translation speed, usually estimated from adjacent track points; used to estimate azimuthal-mean Vmax (Vmaxmean_ms = VmaxNHC_ms - 0.55*Vtrans_ms)
lat = 20  #20 [degN]; default 20N; storm-center latitude
R34ktNHCquadmax_nautmi = 127.04 #(135 + 150 + 145 + 150) / 4 #average NHC R34kt radius (here 4 quadrants)
                                                        # these are officially the MAXIMUM radii of this wind speed in each quadrant;
                                                        # value is reduced by factor 0.85 within the code to estimate the mean radius (see Chavas Knaff Klotzbach 2025 for more info)
Penv_mb = 1008      #[mb]; environmental pressure, to create full pressure profile
## Default values: VmaxNHC_kt=100 kt, R34ktNHCquadmax_nautmi= 145.0 naut mi, lat = 20 --> unadjusted Rmax=38.1 km (sanity check)

# Rmax
Rmax_km = 30     #30 [km]; input value
#Rmax_km = None  #None: estimate Rmax from R34kt -- ref Chavas and Knaff 2022 WAF)
############################################################


################################################################
## Calculate wind and pressure profiles and associated data
"""
Full modeling pipeline:
- If no Rmax input: estimate Rmax from R34kt -- ref Chavas and Knaff 2022 WAF
- Estimate R0 from R34kt: ref Tao et al. (2025, GRL); approximate version of outer model of refs Emanuel 2004 / Chavas et al. 2015 JAS / Chavas and Lin 2016 JAS
- Generate wind profile: Analytic complete wind profile: ref Tao et al. (2025, GRL)
    1) eye: r<Rmax (linear model);
    2) inner-core: Rmax to R34kt (linear-M model; Tao+ 2023 GRL);
    3) intermediate radii: R34kt to transition radius (modified Rankine model; Tao+ 2023 GRL, Klotzbach+ 2022 JGRA); and
    4) large radii: transition radius to outer radius (Ekman suction model; Emanuel 2004; Chavas+ 2015/2016 JAS).
- Estimate Pmin: ref Chavas Knaff Klotzbach 2025 WAF
- Generate pressure profile that matches Pmin: ref Chavas Knaff Klotzbach 2025 WAF
"""
from tcwindprofile.windprofile_all import run_full_wind_model


run_kwargs = dict(
    VmaxNHC_kt=VmaxNHC_kt,
    Vtrans_kt=Vtrans_kt,
    R34kt_quad_max_nautmi=R34ktNHCquadmax_nautmi,
    lat=lat,
    Penv_mb=Penv_mb,
    plot=True,
)

# Only pass Rmax_km if user specified it
if Rmax_km is not None:
    run_kwargs["Rmax_km"] = Rmax_km
    mode_tag = "givenRmax"
else:
    mode_tag = "estRmax"

tc_wind_and_pressure_profile = run_full_wind_model(**run_kwargs)

if Rmax_km is None:
    print(f"[Estimated Rmax] Rmax = {tc_wind_and_pressure_profile['Rmax_km']:.1f} km")
else:
    print(f"[Given Rmax]     Rmax = {tc_wind_and_pressure_profile['Rmax_km']:.1f} km")

print(f"R0 = {tc_wind_and_pressure_profile['R0_km']:.1f} km")
print(f"Pmin = {tc_wind_and_pressure_profile['Pmin_mb']:.1f} hPa")
################################################################

################################################################
################################################################
## Plot that data
from tcwindprofile.plot_windprofile import plot_wind_and_pressure

def plot_profile(tc_profile, save_path):
    # unpack
    rr_km         = tc_profile['rr_km']
    vv_ms         = tc_profile['vv_ms']
    pp_mb         = tc_profile['pp_mb']
    Vmaxmean_ms   = tc_profile['Vmaxmean_ms']
    Rmax_km       = tc_profile['Rmax_km']
    R34ktmean_km  = tc_profile['R34ktmean_km']
    V34kt_ms      = tc_profile['V34kt_ms']
    R0_km         = tc_profile['R0_km']
    Pmin_mb       = tc_profile['Pmin_mb']

    # then:
    Renv_km = R0_km
    plot_wind_and_pressure(
        rr_km, vv_ms,
        Rmax_km, Vmaxmean_ms,
        R34ktmean_km, V34kt_ms,
        R0_km, lat,
        rr_km, pp_mb,
        Renv_km, Penv_mb, Pmin_mb,
        save_path=save_path
    )

plot_profile(tc_wind_and_pressure_profile, save_path=f'tc_wind_pressure_profiles_{mode_tag}')
################################################################



from setuptools import setup, find_packages

setup(
    name='tcwindprofile',
    version='2.1.0',
    packages=find_packages(),
    install_requires=[],
    author='Dan Chavas',
    author_email='drchavas@gmail.com',
    description='Create a fast and robust radial profile of the tropical cyclone rotating wind and pressure field from inputs Vmax, R34kt, latitude, translation speed, and environmental pressure',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/drchavas/tcwindprofile',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    license='MIT',
#    license_files=['LICENSE'],
    python_requires='>=3.6',
)
